//
// ini.hpp:
//	Interface Definition of classes InitFile, Section,
//	Key, AbsPtr, and PointerArray.
//
// Purpose:
//	Reads Windows-style .INI files into memory and provides
//	methods to access and update the values.  For more
//	information, see INI.DOC.
//
// Author(s): Mark Pflaging   C-serve  70233,1552
// Copyright 1992   Mark Thomas Pflaging
//
//
// Date:    5/15/92 - 9/21/92
//
#ifndef __INI_HPP
#define __INI_HPP

#include <fstream.h>

class AbsPtr {
	union {
		void * ptr;
		int index;
	} data;
public:
	AbsPtr(void * whatever = NULL) { data.ptr = whatever; }
	AbsPtr(int whatever) { data.index = whatever; }
	void * get() { return data.ptr; }
	int getInt() { return data.index; }
	void put(void * value) {
		// void * retval = data.ptr;
		data.ptr = value;
		// return retval;
	}
	void put(int value) {
		// int retval = data.index;
		data.index = value;
		// return retval;
	}
	char * putString(char * value);
	int operator++() { return (++(data.index)); }
};

class PointerArray {
	AbsPtr * ptrArray;
	unsigned int numFilled;
	int chunkSize;
	int newElement();
public:
	PointerArray(int chunksize = 10) : numFilled(0),
	    chunkSize(chunksize) {
		ptrArray = new AbsPtr[chunkSize];
	}
	~PointerArray() { dealloc(); }
	void dealloc() {
		delete [] ptrArray;
	}
	AbsPtr &bumpUp(int position, int last);
	int insert(int where, void * element);
	void searchAndInsert(int where,
		int (*compare)(void *, void *), void * element);
	void insertAtEnd(void * value);
	void insertString(char * value);
	int getLastIndex() { return numFilled; }
	AbsPtr * getLastPtr() { return &ptrArray[numFilled]; }
	AbsPtr * getFirstPtr() { return ptrArray; }
	int search(void * element, int lowerBound, int upperBound,
		int (*compare)(void *, void *));
	void * operator[](int index) {
		return (ptrArray[index].get());
	}
	void doAll(void (*func)(AbsPtr & element, void * userPtr),
		void * userPtr);
};

class Key {
	// In this class, 'name' is lval and 'data' is rval.
protected:
	static PointerArray pointers;
	char * name;
	AbsPtr data;
	PointerArray * comments;
	static int compare(void * val1, void * val2);
	AbsPtr & getData() { return data; }
	static Key * maker(char * lval, char * rval) {
		return new Key(lval, rval);
	}
public:
	Key() : name(NULL), comments(NULL) {}
	Key(istream &ini, int where);
	Key(char * lval, char * rval) : comments(NULL) {
		setName(lval);
		data.putString(rval);
	}
	~Key();
	char * nameOf() { return name; }
	char * setName(char * value);
	PointerArray * getComments() { return comments; }
	void clearComments() { comments = NULL; }
	void setComments(PointerArray * value) { comments = value; }
	char * valueOf() { return ((char *)(data.get())); }
	ostream & dump(ostream & where);
};

class Section : public Key {
	// In this class, name is the section name, and
	// data is the index of the first Key in the section.

	static int compare2(void * val1, void * val2);
protected:
	static PointerArray pointers;
	AbsPtr len;	// # of keys in the section.
	PointerArray * insert(istream & ini,
		PointerArray * sectionComments, int where);
	Key * findAndCreate(PointerArray & ptrs,
		int (*compareFunc)(void *, void *), PointerArray & topPtrs,
		Key * (*makeFunc)(char *, char *), char * lval,
		char * rval, int & changed);
	char * find(char * lval, char * rval, int & changed);
	static Key * maker(char * sname, char * dummy) {
		return new Section(sname, dummy);
	}
	void destructor(PointerArray & ptrs,
		void (*killFunc)(void *));
	static void kill(void * arg) {
		Key * temp = (Key *)arg;
		delete((char *)(temp->valueOf()));
		delete temp;
	}
public:
	Section() : len(0) {}
	Section(AbsPtr & aHeading);	// insert a heading in sorted array.
	Section(char * sectionName, char * dummy);
	ostream & dump(ostream & where);
	int getLen() { return (len.getInt()); }
	char * valueOf() {
	    return ("Section::valueOf() - Function not implemented!");
	}
};

class InitFile : public Section {
	// In this class, name is the file name, and
	// data is the index of the first Section in the file.
	// len is # of Sections in the file.
	int changed;

	static PointerArray pointers;
	static int compare(void * val1, void * val2);
	static void kill(void * arg) {
		((InitFile *)arg)->destructor(Key::pointers, Section::kill);
		delete ((Section *)arg);
	}
public:
	InitFile(char * aFile);
	~InitFile() {
		destructor(Section::pointers, kill);
	}
	// dealloc() for TEST PURPOSES ONLY!
	static void dealloc() {
		pointers.dealloc();
		Section::pointers.dealloc();
		Key::pointers.dealloc();
	}
	void write(char * filename = NULL);
	ostream & dump(ostream & where);
	char * find(char * section, char * lval, char *rval = NULL);
	int find(char * section, char * lval, int rval);
};

#endif
